' --------------------------------------------------------------------------------------------------
' elviscpsocket.bas ist ein Basic-Script das in Verbindung mit dem Serial Scriptable Port (elviscpd.exe)
' des Visualisierungssystems Elvis (www.it-gmbh.de) eingesetzt werden kann.
' Es implementiert die Verbindung ber TCP/IP-Socket.
' 
' Diese Schnittstelle implementiert beide Kommunikationsseiten der Socket-Verbindung (Server und Client).
' Alle Einstellungen werden ber die Open-Paramenter aus der Peojektierung vorgenommen.
' Neben dem Parameter fr das Script (Script=<Pfad der script-Datei>) sind die folgenden Parameter mglich:
'
' ServerIP=<IP-Adresse des Servers>			Zwingend beim Client anzugegeben. Bei UDP fr beide Seiten angeben!
' PortID=<Socket ID>						Zwingend bei Server und Client anzugegeben
' ClientKO=0 oder 1						Optional, (client keep open) nur bei Client, 
'														kann angegeben werden wenn die Verbindung offen gehalten werden soll
'														Default: Die Verbindung wird geschlossen, wenn eine Client-Aktion beendet ist.
' Protocol=<protocol type>					Optional, Server und Client, der protocol type kann sein TCP|UDP|ICMP|RAW
'														Default: TCP
' Timer=<Zeitabstand in ms>				Optional,  Server und Client, der Zeitabstand in dem nach empfangegen 
'														Daten gesehen wird.
'														Default: 1000  
'
' Das Trennzeichen zwischen den Parametern ist der "/".
'
'
' ber den Parameter "ServerIP=" wird die IP-Adresse des Servers angegeben mit dem sich der Client 
' verbinden soll.
' Wenn der Parameter "ServerIP=" verwendet wird ist es eine Client-Seite, andernfalls ein Server.
' Der Socket-Server wird stets mit dem Elvis Prozessserver gestartet und mit diesem beendet.
' Der Socket-Client kann entweder fr jede Kommunikationsanforderung eine Verbindung aufbauen (ClientKO=0)
' oder die Verbindung offen lassen (ClientKO=1).
'
' Beispiel-Parameter des Elvis CustomPortDriver fr die Server-Seite:
' Script=C:\Programme\Elvis\ElvisCpd\elviscpsocket.bas;OpenParam="PortID=7002"
'
' Beispiel-Parameter des Elvis CustomPortDriver fr die Client-Seite:
' Script=C:\Programme\Elvis\ElvisCpd\elviscpsocket.bas;OpenParam="ServerIP=192.168.255.211/PortID=7002/ClientKO=1"
'
' Fr die Verbindung zwischen zwei Elvis-Servern ist ein einfaches Protokoll definiert:
'
' Telegrammaufbau: <Telegramm-Typ><Adresse>@<Wert>
' Der Typ ist ein Zeichen lang und dient zur unterscheidung ob ein Wert geschrieben, abgefragt 
' oder besttigt wird. Siehe unten.
' Die Adresse und der Wert werden direkt aus Elvis bernommen und in das Telegramm eingefgt.
' Das Besttigen erfolgt durch zurcksenden des gesamten Telegramms jedoch mit dem Besttigungs-Zeichen.
'
' --------------------------------------------------------------------------------------------------
' Historie:
' 31.01.2006, Harald Pokorny:  Makro erstellt
' 13.02.2006, Harald Pokorny:  UDP
' Copyright (C) 2006, IT GmbH
' --------------------------------------------------------------------------------------------------
' Dieses Makro ist Teil des Softwareprodukts Elvis. Es kann frei verwendet,
' modifiziert und verteilt werden, vorausgesetzt, obige Copyright-Notiz wird nicht entfernt.
' Ein Gewhrleistung fr die korrekte Funktion des Makros wird nicht bernommen.
' --------------------------------------------------------------------------------------------------


Option Explicit
Option Base 0
Global g_Server As Boolean					'wenn true -> es ist ein Server, sonst ein Client
Global g_ClientKeepOpen As Boolean		'wenn true -> der Client hlt die Verbindung offen
Global g_ConnectionOpen As Boolean
Global g_ServerIP As String
Global g_PortID As String
Global g_Protocol As String
Global g_Timer As String


Const PARAM_SERVERIP = "ServerIP="
Const PARAM_PORTID = "PortID="
Const PARAM_CLIENT_KEEPOPEN = "ClientKO="
Const PARAM_PROTOCOL = "Protocol="
Const PARAM_TIMER = "Timer="

'Protokolldefinition
Const SEPERATOR_SIGN = "@"			'Zeichen zwischen Adresse (dst) und Wert (value)
Const REQUEST_SIGN = "?"			'Zeichen zum Abfragen des Wertes
Const SEND_SIGN = "!"				'Zeichen zum Senden eines Wertes (Wert setzen)
Const ACKNOWLEDGE_SIGN = "*"		'Zeichen dass das Telegramm ein Acknowlede Telegramm ist

' -------------------------------------------------------
' Socket definitions and prototypes
Sub OnOpen(sParam As String)
	g_ConnectionOpen = False
	Print "Die Parameter sind: " & sParam
	
	g_ServerIP = GetOpenParam(sParam, PARAM_SERVERIP, "")
	If CheckIP(g_ServerIP) Then 
		g_Server = False
	Else
		g_Server = True
	End If
	
	g_PortID  = GetOpenParam(sParam, PARAM_PORTID, "")
	g_ClientKeepOpen = CBool(GetOpenParam(sParam, PARAM_CLIENT_KEEPOPEN, "0"))	
	g_Protocol = GetOpenParam(sParam, PARAM_PROTOCOL, "TCP")
	g_Timer = GetOpenParam(sParam, PARAM_TIMER, "1000")
			
	Print "g_ServerIP = " & g_ServerIP
	Print "Server = " & g_Server
	Print "g_PortID = " & g_PortID
	Print "ClientKeepOpen = " &  g_ClientKeepOpen
	Print "g_Protocol = " & g_Protocol
	Print "g_Timer = " & g_Timer 
	
	Open_Internal
End Sub

Sub OnClose()
    Close_Internal
End Sub


Sub Open_Internal()
	Dim sParam As String
	
	If g_Protocol = "TCP" Then
	
		If g_Server Then
			If Not g_ConnectionOpen Then
				'sParam = "Port=" & g_PortID & ";Protocol=" & g_Protocol & ";Timer= " & g_Timer 
				sParam = "Port=" & g_PortID & ";Protocol=" & g_Protocol '& ";Timer= " & g_Timer 
				'sParam = "Port=" & g_PortID

				Print "Try to open server, parameter: " & sParam
				CreateSocketServer sParam 	
				g_ConnectionOpen = True
			End If
		Else
			If Not g_ConnectionOpen Then
				'sParam = """Protocol=" & CStr(g_Protocol) & ";Timer= " & CStr(g_Timer) & """"
				sParam = "Port=" & g_PortID & ";Protocol=" & g_Protocol '& ";Timer= " & g_Timer 
				'sParam = """Protocol=" & CStr(g_Protocol) & """"
				Print "Try to open client, parameter: " & sParam
				ConnectSocketClient g_ServerIP, CInt(g_PortID), sParam
				g_ConnectionOpen = True
			End If
		End If
	ElseIf g_Protocol = "UDP" Then

			If Not g_ConnectionOpen Then
				'sParam = "Port=" & g_PortID & ";Protocol=" & g_Protocol & ";Timer= " & g_Timer 
				sParam = "Port=" & g_PortID & ";Protocol=" & g_Protocol '& ";Timer= " & g_Timer 
				'sParam = "Port=" & g_PortID

				Print "Try to open UDP server, parameter: " & sParam
				CreateSocketServer sParam 	
				g_ConnectionOpen = True
			End If
				
	Else
    		Print "Open of ICMP|RAW or other protocols are not implemented"
    	End If
    
End Sub

Sub Close_Internal()
	If g_Protocol = "TCP" Then
		If g_Server Then
			'DisconnectSocketServer
    		Else
 	   		If g_ClientKeepOpen Then
  	  		'Verbindung offen lassen
   	 		Else
    			    DisconnectSocketClient
    			End If
    		End If
	ElseIf g_Protocol = "UDP" Then
	'do nothing
	
    	End If
End Sub

Function GetOpenParam(sParam As String, sIdentifier As String, sDef As String) As String
	Dim iPos As Integer
	Dim iSep As Integer
	Dim sValue As String
	
	If Left(sParam, 1) = """" Then 	sParam = Right(sParam, Len(sParam)-1)
	If Right(sParam, 1) = """" Then sParam = Left(sParam, Len(sParam)-1)
	
	iPos = InStr(1, sParam, sIdentifier)
	If iPos > 0 Then
		iPos = iPos + Len(sIdentifier)
		iSep = InStr(iPos, sParam, "/")
		If iSep > 0 Then
			sValue = Mid(sParam, iPos, iSep-iPos)
		Else
			sValue = Right(sParam, Len(sParam) - iPos + 1)
		End If
	Else
	'default
	sValue = sDef		
	End If	
	Print "sParam: " & sParam & "  sValue: " & sValue
	GetOpenParam = sValue
End Function

Function CheckIP(ByVal sIP As String) As Boolean
	Dim nPos As Integer
	Dim sIPmsg As String
	Dim i As Integer
	
	CheckIP = True
	sIPmsg = "elviscpsocket - wrong IP: " & sIP
	For i = 1 To 4
		If i < 4 Then
			nPos = InStr(1,sIP,".")
			If nPos < 1 Or nPos > 4 Then GoTo ExitError
			If Not IsNumeric(Left(sIP,nPos - 1)) Then GoTo ExitError
			sIP = Right(sIP, Len(sIP) - nPos)	
		Else
			If Not IsNumeric(sIP) Then GoTo ExitError
		End If
	Next i
	If False Then
ExitError:
		CheckIP = False
		Print sIPmsg
	End If
End Function

Sub Send_Internal(sSend As String)
	Dim hRes As Integer
	Dim cbMsg As Integer
	Dim i As Integer
	Print "Sending: " & sSend 
	Open_Internal
    
	cbMsg = Len(sSend)
	Dim vMsgToSend() As Byte
	ReDim vMsgToSend(cbMsg - 1)
	For i = 0 To cbMsg - 1
		vMsgToSend(i) = Asc(Mid(sSend, 1 + i, 1))
	Next i
	If g_Protocol = "TCP" Then
		hRes = SocketSend(vMsgToSend)
	ElseIf g_Protocol = "UDP" Then
		hRes = SocketSendTo(g_ServerIP, g_PortID ,vMsgToSend)
	Else
    		Print "Send via ICMP|RAW or other protocols are not implemented"
	End If
	If hRes Then Print "elviscpsocket - Send: hRes=" & CStr(hRes)
	Close_Internal
End Sub

Sub SendValue(dst As String, value As Variant, svrinfo As String)
	'Print "SendValue " & dst & " : " & value
	Send_Internal SEND_SIGN & dst & SEPERATOR_SIGN & CStr(value)
End Sub


Function Date_Time()
    Dim now_ As Date
    now_ = Now()
    Date_Time = Format(now_, "ddmmyy") & " " & Format(now_, "hhnnss")
End Function

Sub OnReceive(ByVal sData As String)
	Print "Receive: " & sData
	Dim sType As String
	Dim sAddr As String
	Dim sValue As String
	Dim iPos As Integer
	iPos = InStr(sData, SEPERATOR_SIGN)
	sType = Left(sData, 1)
	sAddr = Mid(sData, 2, iPos-2)
	sValue = Right(sData, Len(sData)-iPos)
	
	Print "iPos=" & CStr(iPos) & "  sType: " & sType & "  sAddr: " & sAddr & "  sValue: " & sValue


	
	Select Case sType
		Case SEND_SIGN
			Open_Internal
			FireReceiveValue sAddr, sValue
			Send_Internal ACKNOWLEDGE_SIGN & sAddr & SEPERATOR_SIGN & sValue		'als Empfangsbesttigung wird das Telegramm zurckgesendet. 		
			Close_Internal
		Case REQUEST_SIGN
			'Auf Anfrage antworten - kann Elvis Server nicht (darf er auch nicht knnen)
			Print "elviscpsocket - Request an Elvis Server nicht mglich."	
		Case ACKNOWLEDGE_SIGN
			Open_Internal
			FireReceiveValue sAddr, sValue
			Close_Internal
		Case Else
	End Select
End Sub

Sub OnReceiveFrom(ByVal sAddr As String, ByVal nPort As Long, ByVal sData As String)
	Print "ReceiveFrom: " & sAddr & " on Port: " & CStr(nPort) & " Data: "& sData
	OnReceive sData
End Sub


Sub RequestValue(dst As String, svrinfo As String)
    'Der Elvis Server ist bei der Kommunikation mit Fremdsystemen immer in 
    'der Rolle eines Clients der Daten von einer Anlage / dem Fremdsystem erhlt.
    'Er kann also nicht abgefragt werden. Wenn die Gegenstelle also auch ein 
    'Elvis Server ist, ist die Abfrage sinnlos.
    
    Send_Internal REQUEST_SIGN & dst & SEPERATOR_SIGN
End Sub


